import unittest

import wrapt

from compat import PYXY


@wrapt.synchronized
def function():
    print("function")


class C1:

    @wrapt.synchronized
    def function1(self):
        print("function1")

    @wrapt.synchronized
    @classmethod
    def function2(cls):
        print("function2")

    @wrapt.synchronized
    @staticmethod
    def function3():
        print("function3")


c1 = C1()


@wrapt.synchronized
class C2:
    pass


@wrapt.synchronized
class C3:
    pass


class C4:

    # Prior to Python 3.9, this yields undesirable results due to how
    # class method is implemented. The classmethod doesn't bind the
    # method to the class before calling. As a consequence, the
    # decorator wrapper function sees the instance as None with the
    # class being explicitly passed as the first argument. It isn't
    # possible to detect and correct this. For more details see:
    # https://bugs.python.org/issue19072

    @classmethod
    @wrapt.synchronized
    def function2(cls):
        print("function2")

    @staticmethod
    @wrapt.synchronized
    def function3():
        print("function3")


c4 = C4()


class C5:

    def __bool__(self):
        return False

    __nonzero__ = __bool__

    @wrapt.synchronized
    def function1(self):
        print("function1")


c5 = C5()


class TestSynchronized(unittest.TestCase):

    def test_synchronized_function(self):
        _lock0 = getattr(function, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        function()

        _lock1 = getattr(function, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        function()

        _lock2 = getattr(function, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        function()

        _lock3 = getattr(function, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_inner_staticmethod(self):
        _lock0 = getattr(C1.function3, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c1.function3()

        _lock1 = getattr(C1.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        C1.function3()

        _lock2 = getattr(C1.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        C1.function3()

        _lock3 = getattr(C1.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_outer_staticmethod(self):
        _lock0 = getattr(C4.function3, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c4.function3()

        _lock1 = getattr(C4.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        C4.function3()

        _lock2 = getattr(C4.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        C4.function3()

        _lock3 = getattr(C4.function3, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_inner_classmethod(self):
        if hasattr(C1, "_synchronized_lock"):
            del C1._synchronized_lock

        _lock0 = getattr(C1, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c1.function2()

        _lock1 = getattr(C1, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        C1.function2()

        _lock2 = getattr(C1, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        C1.function2()

        _lock3 = getattr(C1, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_outer_classmethod(self):
        # Prior to Python 3.9 this isn't detected as a class method
        # call, as the classmethod decorator doesn't bind the wrapped
        # function to the class before calling and just calls it direct,
        # explicitly passing the class as first argument. For more
        # details see: https://bugs.python.org/issue19072
        # Starting with Python 3.13 the old behavior is back.
        # For more details see https://github.com/python/cpython/issues/89519

        if (3, 9) <= PYXY < (3, 13):
            _lock0 = getattr(C4, "_synchronized_lock", None)
        else:
            _lock0 = getattr(C4.function2, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c4.function2()

        if (3, 9) <= PYXY < (3, 13):
            _lock1 = getattr(C4, "_synchronized_lock", None)
        else:
            _lock1 = getattr(C4.function2, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        C4.function2()

        if (3, 9) <= PYXY < (3, 13):
            _lock2 = getattr(C4, "_synchronized_lock", None)
        else:
            _lock2 = getattr(C4.function2, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        C4.function2()

        if (3, 9) <= PYXY < (3, 13):
            _lock3 = getattr(C4, "_synchronized_lock", None)
        else:
            _lock3 = getattr(C4.function2, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_instancemethod(self):
        if hasattr(C1, "_synchronized_lock"):
            del C1._synchronized_lock

        _lock0 = getattr(c1, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        C1.function1(c1)

        _lock1 = getattr(c1, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        c1.function1()

        _lock2 = getattr(c1, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        c1.function1()

        _lock3 = getattr(c1, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

        del c1._synchronized_lock

        C1.function2()

        _lock4 = getattr(C1, "_synchronized_lock", None)
        self.assertNotEqual(_lock4, None)

        c1.function1()

        _lock5 = getattr(c1, "_synchronized_lock", None)
        self.assertNotEqual(_lock5, None)
        self.assertNotEqual(_lock5, _lock4)

    def test_synchronized_type_new_style(self):
        if hasattr(C2, "_synchronized_lock"):
            del C2._synchronized_lock

        _lock0 = getattr(C2, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c2 = C2()

        _lock1 = getattr(C2, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        c2 = C2()

        _lock2 = getattr(C2, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        c2 = C2()

        _lock3 = getattr(C2, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_type_old_style(self):
        if hasattr(C3, "_synchronized_lock"):
            del C3._synchronized_lock

        _lock0 = getattr(C3, "_synchronized_lock", None)
        self.assertEqual(_lock0, None)

        c2 = C3()

        _lock1 = getattr(C3, "_synchronized_lock", None)
        self.assertNotEqual(_lock1, None)

        c2 = C3()

        _lock2 = getattr(C3, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)
        self.assertEqual(_lock2, _lock1)

        c2 = C3()

        _lock3 = getattr(C3, "_synchronized_lock", None)
        self.assertNotEqual(_lock3, None)
        self.assertEqual(_lock3, _lock2)

    def test_synchronized_false_instance(self):
        c5.function1()

        self.assertEqual(bool(c5), False)

        _lock1 = getattr(C5, "_synchronized_lock", None)
        self.assertEqual(_lock1, None)

        _lock2 = getattr(c5, "_synchronized_lock", None)
        self.assertNotEqual(_lock2, None)


if __name__ == "__main__":
    unittest.main()



import unittest

import wrapt


class TestPartialCallableObjectProxy(unittest.TestCase):

    def test_no_arguments(self):
        def func0():
            return ((), {})

        partial0 = wrapt.partial(func0)

        args, kwargs = (), {}

        self.assertEqual(partial0(), (args, kwargs))

    def test_empty_arguments(self):
        def func0(*args, **kwargs):
            return (args, kwargs)

        args, kwargs = (), {}

        partial0 = wrapt.partial(func0, *args, **kwargs)

        self.assertEqual(partial0(), (args, kwargs))

    def test_1_positional_argument(self):
        def func0(*args, **kwargs):
            return (args, kwargs)

        args, kwargs = (1,), {}

        partial0 = wrapt.partial(func0, *args)

        self.assertEqual(partial0(), (args, kwargs))

    def test_1_keyword_argument(self):
        def func0(*args, **kwargs):
            return (args, kwargs)

        args, kwargs = (), {"k1": 1}

        partial0 = wrapt.partial(func0, **kwargs)

        self.assertEqual(partial0(), (args, kwargs))

    def test_multiple_positional_arguments(self):
        def func0(*args, **kwargs):
            return (args, kwargs)

        args, kwargs = (1, 2, 3), {}

        partial0 = wrapt.partial(func0, *args)

        self.assertEqual(partial0(), (args, kwargs))

    def test_multiple_keyword_arguments(self):
        def func0(*args, **kwargs):
            return (args, kwargs)

        args, kwargs = (), {"k1": 1, "k2": 2, "k3": 3}

        partial0 = wrapt.partial(func0, **kwargs)

        self.assertEqual(partial0(), (args, kwargs))


if __name__ == "__main__":
    unittest.main()



"""
This example demonstrates usage of the decorator() function.
"""

from typing import Any, Callable, ParamSpec, TypeVar

import wrapt

P = ParamSpec("P")
R = TypeVar("R")


@wrapt.decorator
def wrapper1(
    wrapped: Callable[[int], int],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> int:
    return wrapped(*args, **kwargs)


@wrapper1
def function1(x: int) -> int:
    """A function that takes an integer and returns it."""
    return x


result1a: int = function1(1)

# Mismatch return type. (FAIL)
result1b: str = function1(1)

# No arguments provided. (FAIL)
function1()

function1(1)

# Provide incorrect argument type. (FAIL)
function1("x")

# Wrong number of arguments provided. (FAIL)
function1(1, 2)


@wrapt.decorator(enabled=True)
def wrapper2(
    wrapped: Callable[[int], int],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> int:
    return wrapped(*args, **kwargs)


@wrapper2
def function2(x: int) -> int:
    """A function that takes an integer and returns it."""
    return x


# No arguments provided. (FAIL)
function2()

function2(1)

# Provide incorrect argument type. (FAIL)
function2("x")

# Wrong number of arguments provided. (FAIL)
function2(1, 2)


def adapter1(i: int) -> str:
    return str(i)


def wrapper3(wrapped: Callable[[int], int]) -> Callable[[int], str]:
    @wrapt.decorator(adapter=adapter1)
    def _wrapper3(
        wrapped: Callable[..., Any],
        instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        return adapter1(wrapped(*args, **kwargs))

    return _wrapper3(wrapped)


@wrapper3
def function3(x: int) -> int:
    """A function that takes an integer and returns it."""
    return x


result3a: str = function3(1)

# Mismatch return type. (FAIL)
result3b: int = function3(1)

# No arguments provided. (FAIL)
function3()

function3(1)

# Provide incorrect argument type. (FAIL)
function3("x")

# Wrong number of arguments provided. (FAIL)
function3(1, 2)


@wrapt.decorator(adapter=adapter1)
def wrapper4(
    wrapped: Callable[[int], int],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return adapter1(wrapped(*args, **kwargs))


@wrapper4
def function4(x: int) -> int:
    """A function that takes an integer and returns it."""
    return x


result4a: str = function4(1)

# Allowed since return Any.
result4b: int = function4(1)

# No arguments provided. (FAIL)
function4()

function4(1)

# Provide incorrect argument type. (FAIL)
function4("x")

# Wrong number of arguments provided. (FAIL)
function4(1, 2)


def adapter2(a: int, b: int) -> str:
    return str(a + b)


def wrapper5(wrapped: Callable[[int], int]) -> Callable[[int, int], str]:
    @wrapt.decorator(adapter=adapter2)
    def _wrapper5(
        wrapped: Callable[..., Any],
        instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> Any:
        def bind(a: int, b: int) -> tuple[int, int]:
            return a, b

        a, b = bind(*args, **kwargs)

        return str(wrapped(a + b))

    return _wrapper5(wrapped)


@wrapper5
def function5(x: int) -> int:
    """A function that takes an integer and returns it."""
    return x


result5: str = function5(1, 2)


class Example:
    @wrapt.decorator
    def pass_through_im(
        self: Any,
        wrapped: Callable[P, R],
        instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> R:
        return wrapped(*args, **kwargs)

    @wrapt.decorator
    @classmethod
    def pass_through_cm(
        cls: Any,
        wrapped: Callable[P, R],
        instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> R:
        return wrapped(*args, **kwargs)

    @wrapt.decorator
    @staticmethod
    def pass_through_sm(
        wrapped: Callable[P, R],
        instance: Any,
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> R:
        return wrapped(*args, **kwargs)


example_instance = Example()


@example_instance.pass_through_im
def function6(a: int, b: int = 0) -> int:
    return a + b


result6: int = function6(1, 2)

function6(1)
function6(1, b=2)


@example_instance.pass_through_cm
def function7(a: int, b: int = 0) -> int:
    return a + b


result7: int = function7(1, 2)

function7(1)
function7(1, b=2)


@example_instance.pass_through_sm
def function8(a: int, b: int = 0) -> int:
    return a + b


result8: int = function8(1, 2)

function8(1)
function8(1, b=2)



"""Variants of ObjectProxy for different use cases."""

from collections.abc import Callable
from types import ModuleType

from .__wrapt__ import BaseObjectProxy
from .decorators import synchronized

# Define ObjectProxy which for compatibility adds `__iter__()` support which
# has been removed from `BaseObjectProxy`.


class ObjectProxy(BaseObjectProxy):
    """A generic object proxy which forwards special methods as needed.
    For backwards compatibility this class adds support for `__iter__()`. If
    you don't need backward compatibility for `__iter__()` support then it is
    preferable to use `BaseObjectProxy` directly. If you want automatic
    support for special dunder methods for callables, iterators, and async,
    then use `AutoObjectProxy`."""

    @property
    def __object_proxy__(self):
        return ObjectProxy

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __iter__(self):
        return iter(self.__wrapped__)


# Define variant of ObjectProxy which can automatically adjust to the wrapped
# object and add special dunder methods.


def __wrapper_call__(*args, **kwargs):
    def _unpack_self(self, *args):
        return self, args

    self, args = _unpack_self(*args)

    return self.__wrapped__(*args, **kwargs)


def __wrapper_iter__(self):
    return iter(self.__wrapped__)


def __wrapper_next__(self):
    return self.__wrapped__.__next__()


def __wrapper_aiter__(self):
    return self.__wrapped__.__aiter__()


async def __wrapper_anext__(self):
    return await self.__wrapped__.__anext__()


def __wrapper_length_hint__(self):
    return self.__wrapped__.__length_hint__()


def __wrapper_await__(self):
    return (yield from self.__wrapped__.__await__())


def __wrapper_get__(self, instance, owner):
    return self.__wrapped__.__get__(instance, owner)


def __wrapper_set__(self, instance, value):
    return self.__wrapped__.__set__(instance, value)


def __wrapper_delete__(self, instance):
    return self.__wrapped__.__delete__(instance)


def __wrapper_set_name__(self, owner, name):
    return self.__wrapped__.__set_name__(owner, name)


class AutoObjectProxy(BaseObjectProxy):
    """An object proxy which can automatically adjust to the wrapped object
    and add special dunder methods as needed. Note that this creates a new
    class for each instance, so it has much higher memory overhead than using
    `BaseObjectProxy` directly. If you know what special dunder methods you need
    then it is preferable to use `BaseObjectProxy` directly and add them to a
    subclass as needed. If you only need `__iter__()` support for backwards
    compatibility then use `ObjectProxy` instead.
    """

    def __new__(cls, wrapped):
        """Injects special dunder methods into a dynamically created subclass
        as needed based on the wrapped object.
        """

        namespace = {}

        wrapped_attrs = dir(wrapped)
        class_attrs = set(dir(cls))

        if callable(wrapped) and "__call__" not in class_attrs:
            namespace["__call__"] = __wrapper_call__

        if "__iter__" in wrapped_attrs and "__iter__" not in class_attrs:
            namespace["__iter__"] = __wrapper_iter__

        if "__next__" in wrapped_attrs and "__next__" not in class_attrs:
            namespace["__next__"] = __wrapper_next__

        if "__aiter__" in wrapped_attrs and "__aiter__" not in class_attrs:
            namespace["__aiter__"] = __wrapper_aiter__

        if "__anext__" in wrapped_attrs and "__anext__" not in class_attrs:
            namespace["__anext__"] = __wrapper_anext__

        if "__length_hint__" in wrapped_attrs and "__length_hint__" not in class_attrs:
            namespace["__length_hint__"] = __wrapper_length_hint__

        # Note that not providing compatibility with generator-based coroutines
        # (PEP 342) here as they are removed in Python 3.11+ and were deprecated
        # in 3.8.

        if "__await__" in wrapped_attrs and "__await__" not in class_attrs:
            namespace["__await__"] = __wrapper_await__

        if "__get__" in wrapped_attrs and "__get__" not in class_attrs:
            namespace["__get__"] = __wrapper_get__

        if "__set__" in wrapped_attrs and "__set__" not in class_attrs:
            namespace["__set__"] = __wrapper_set__

        if "__delete__" in wrapped_attrs and "__delete__" not in class_attrs:
            namespace["__delete__"] = __wrapper_delete__

        if "__set_name__" in wrapped_attrs and "__set_name__" not in class_attrs:
            namespace["__set_name__"] = __wrapper_set_name__

        name = cls.__name__

        if cls is AutoObjectProxy:
            name = BaseObjectProxy.__name__

        return super(AutoObjectProxy, cls).__new__(type(name, (cls,), namespace))

    def __wrapped_setattr_fixups__(self):
        """Adjusts special dunder methods on the class as needed based on the
        wrapped object, when `__wrapped__` is changed.
        """

        cls = type(self)
        class_attrs = set(dir(cls))

        if callable(self.__wrapped__):
            if "__call__" not in class_attrs:
                cls.__call__ = __wrapper_call__
        elif getattr(cls, "__call__", None) is __wrapper_call__:
            delattr(cls, "__call__")

        if hasattr(self.__wrapped__, "__iter__"):
            if "__iter__" not in class_attrs:
                cls.__iter__ = __wrapper_iter__
        elif getattr(cls, "__iter__", None) is __wrapper_iter__:
            delattr(cls, "__iter__")

        if hasattr(self.__wrapped__, "__next__"):
            if "__next__" not in class_attrs:
                cls.__next__ = __wrapper_next__
        elif getattr(cls, "__next__", None) is __wrapper_next__:
            delattr(cls, "__next__")

        if hasattr(self.__wrapped__, "__aiter__"):
            if "__aiter__" not in class_attrs:
                cls.__aiter__ = __wrapper_aiter__
        elif getattr(cls, "__aiter__", None) is __wrapper_aiter__:
            delattr(cls, "__aiter__")

        if hasattr(self.__wrapped__, "__anext__"):
            if "__anext__" not in class_attrs:
                cls.__anext__ = __wrapper_anext__
        elif getattr(cls, "__anext__", None) is __wrapper_anext__:
            delattr(cls, "__anext__")

        if hasattr(self.__wrapped__, "__length_hint__"):
            if "__length_hint__" not in class_attrs:
                cls.__length_hint__ = __wrapper_length_hint__
        elif getattr(cls, "__length_hint__", None) is __wrapper_length_hint__:
            delattr(cls, "__length_hint__")

        if hasattr(self.__wrapped__, "__await__"):
            if "__await__" not in class_attrs:
                cls.__await__ = __wrapper_await__
        elif getattr(cls, "__await__", None) is __wrapper_await__:
            delattr(cls, "__await__")

        if hasattr(self.__wrapped__, "__get__"):
            if "__get__" not in class_attrs:
                cls.__get__ = __wrapper_get__
        elif getattr(cls, "__get__", None) is __wrapper_get__:
            delattr(cls, "__get__")

        if hasattr(self.__wrapped__, "__set__"):
            if "__set__" not in class_attrs:
                cls.__set__ = __wrapper_set__
        elif getattr(cls, "__set__", None) is __wrapper_set__:
            delattr(cls, "__set__")

        if hasattr(self.__wrapped__, "__delete__"):
            if "__delete__" not in class_attrs:
                cls.__delete__ = __wrapper_delete__
        elif getattr(cls, "__delete__", None) is __wrapper_delete__:
            delattr(cls, "__delete__")

        if hasattr(self.__wrapped__, "__set_name__"):
            if "__set_name__" not in class_attrs:
                cls.__set_name__ = __wrapper_set_name__
        elif getattr(cls, "__set_name__", None) is __wrapper_set_name__:
            delattr(cls, "__set_name__")


class LazyObjectProxy(AutoObjectProxy):
    """An object proxy which can generate/create the wrapped object on demand
    when it is first needed.
    """

    def __new__(cls, callback=None, *, interface=...):
        """Injects special dunder methods into a dynamically created subclass
        as needed based on the wrapped object.
        """

        if interface is ...:
            interface = type(None)

        namespace = {}

        interface_attrs = dir(interface)
        class_attrs = set(dir(cls))

        if "__call__" in interface_attrs and "__call__" not in class_attrs:
            namespace["__call__"] = __wrapper_call__

        if "__iter__" in interface_attrs and "__iter__" not in class_attrs:
            namespace["__iter__"] = __wrapper_iter__

        if "__next__" in interface_attrs and "__next__" not in class_attrs:
            namespace["__next__"] = __wrapper_next__

        if "__aiter__" in interface_attrs and "__aiter__" not in class_attrs:
            namespace["__aiter__"] = __wrapper_aiter__

        if "__anext__" in interface_attrs and "__anext__" not in class_attrs:
            namespace["__anext__"] = __wrapper_anext__

        if (
            "__length_hint__" in interface_attrs
            and "__length_hint__" not in class_attrs
        ):
            namespace["__length_hint__"] = __wrapper_length_hint__

        # Note that not providing compatibility with generator-based coroutines
        # (PEP 342) here as they are removed in Python 3.11+ and were deprecated
        # in 3.8.

        if "__await__" in interface_attrs and "__await__" not in class_attrs:
            namespace["__await__"] = __wrapper_await__

        if "__get__" in interface_attrs and "__get__" not in class_attrs:
            namespace["__get__"] = __wrapper_get__

        if "__set__" in interface_attrs and "__set__" not in class_attrs:
            namespace["__set__"] = __wrapper_set__

        if "__delete__" in interface_attrs and "__delete__" not in class_attrs:
            namespace["__delete__"] = __wrapper_delete__

        if "__set_name__" in interface_attrs and "__set_name__" not in class_attrs:
            namespace["__set_name__"] = __wrapper_set_name__

        name = cls.__name__

        return super(AutoObjectProxy, cls).__new__(type(name, (cls,), namespace))

    def __init__(self, callback=None, *, interface=...):
        """Initialize the object proxy with wrapped object as `None` but due
        to presence of special `__wrapped_factory__` attribute addded first,
        this will actually trigger the deferred creation of the wrapped object
        when first needed.
        """

        if callback is not None:
            self.__wrapped_factory__ = callback

        super().__init__(None)

    __wrapped_initialized__ = False

    def __wrapped_factory__(self):
        return None

    def __wrapped_get__(self):
        """Gets the wrapped object, creating it if necessary."""

        # We synchronize on the class type, which will be unique to this instance
        # since we inherit from `AutoObjectProxy` which creates a new class
        # for each instance. If we synchronize on `self` or the method then
        # we can end up in infinite recursion via `__getattr__()`.

        with synchronized(type(self)):
            # We were called because `__wrapped__` was not set, but because of
            # multiple threads we may find that it has been set by the time
            # we get the lock. So check again now whether `__wrapped__` is set.
            # If it is then just return it, otherwise call the factory to
            # create it.

            if self.__wrapped_initialized__:
                return self.__wrapped__

            self.__wrapped__ = self.__wrapped_factory__()

            self.__wrapped_initialized__ = True

            return self.__wrapped__


def lazy_import(name, attribute=None, *, interface=...):
    """Lazily imports the module `name`, returning a `LazyObjectProxy` which
    will import the module when it is first needed. When `name is a dotted name,
    then the full dotted name is imported and the last module is taken as the
    target. If `attribute` is provided then it is used to retrieve an attribute
    from the module.
    """

    if attribute is not None:
        if interface is ...:
            interface = Callable
    else:
        if interface is ...:
            interface = ModuleType

    def _import():
        module = __import__(name, fromlist=[""])

        if attribute is not None:
            return getattr(module, attribute)

        return module

    return LazyObjectProxy(_import, interface=interface)



"""
This example demonstrates the incorrect usage of the FunctionWrapper type.

It covers the following cases:
- Wrapper as non callable object
- Wrapper as None object

Should fail mypy type checking for incorrect cases.
"""

from wrapt import FunctionWrapper


def function(x: int, y: str = "default") -> str:
    """A simple function to be wrapped."""
    return f"{x}: {y}"


wrapped_function = FunctionWrapper(function, "string")
wrapped_function = FunctionWrapper(function, None)



import os
import pathlib
import re
import sys
import warnings

import pytest

try:
    from pytest import File as FileCollector
except ImportError:
    from pytest.collect import File as FileCollector

version = tuple(sys.version_info[:2])


# Set MYPYPATH to the src directory relative to this conftest.py file. This
# allows mypy to find the source code for type checking.
_conftest_dir = pathlib.Path(__file__).parent
_src_dir = _conftest_dir.parent / "src"
os.environ["MYPYPATH"] = str(_src_dir)


class DummyCollector(FileCollector):
    def collect(self):
        return []


def construct_dummy(path, parent):
    if hasattr(DummyCollector, "from_parent"):
        item = DummyCollector.from_parent(parent, path=path)
        return item
    else:
        return DummyCollector(path, parent=parent)


def pytest_pycollect_makemodule(module_path, parent):
    basename = module_path.name

    # Handle Python 2/3 general cases
    if "_py2" in basename and version >= (3, 0):
        return construct_dummy(module_path, parent)
    if "_py3" in basename and version < (3, 0):
        return construct_dummy(module_path, parent)

    # Handle specific Python version cases using regex
    # Match patterns like "_py33", "_py34", "_py310", etc.
    version_match = re.search(r"_py(\d)(\d*)", basename)
    if version_match:
        major = int(version_match.group(1))
        minor_str = version_match.group(2)
        minor = int(minor_str) if minor_str else 0

        # Check if current version is less than the required version
        if version < (major, minor):
            return construct_dummy(module_path, parent)

    return None


# -----------------------------
# Custom mypy_*.py + .out tests
# -----------------------------


def run_custom_action(py_file: pathlib.Path) -> str:
    """
    Run mypy on the given file with the current interpreter's major.minor version
    and return the combined stdout/stderr output as text.
    """
    import platform
    import subprocess

    major, minor = version
    cmd = [
        "mypy",
        "--strict",
        "--show-error-codes",
        "--python-version",
        f"{major}.{minor}",
        str(py_file),
    ]
    proc = subprocess.run(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
        check=False,
    )
    output = proc.stdout

    # On Windows, convert backslash paths to forward slashes for consistency
    # with .out files
    if platform.system() == "Windows":
        output = re.sub(r"\btests\\mypy\\", "tests/mypy/", output)

    return output


class MypyPairItem(pytest.Item):
    def __init__(self, name, parent, py_path: pathlib.Path, out_path: pathlib.Path):
        super().__init__(name, parent)
        self.py_path = py_path
        self.out_path = out_path

    def runtest(self):
        # Try to run mypy; if it's not found, skip the test
        try:
            actual_output = run_custom_action(self.py_path)
        except FileNotFoundError:
            warnings.warn(
                f"mypy not found; skipping test {self.py_path.name}",
                UserWarning,
                stacklevel=2,
            )
            pytest.skip("mypy command not found")

        expected_output = self.out_path.read_text(encoding="utf-8")

        # Normalize line endings to avoid platform discrepancies
        if actual_output.replace("\r\n", "\n") != expected_output.replace("\r\n", "\n"):
            raise AssertionError(
                f"Output did not match expected for {self.py_path.name}\n"
                f"Expected (from {self.out_path.name}):\n{expected_output}\n"
                f"Actual:\n{actual_output}"
            )

    def reportinfo(self):
        return self.py_path, 0, f"mypy-pair: {self.py_path.name}"


class MypyPairCollector(pytest.File):
    """
    A collector that discovers mypy*.py files with corresponding .out files in
    the same directory and creates test items for them.
    """

    def collect(self):
        # Only run this custom collection on Python 3.10+
        if version < (3, 10):
            return

        # Skip mypy tests if running on PyPy
        if sys.implementation.name == "pypy":
            return

        path = pathlib.Path(str(self.fspath))
        # Only operate in a tests directory context
        if path.name != "conftest.py":
            return

        tests_dir = path.parent

        for py_file in sorted(tests_dir.glob("mypy/mypy_*.py")):
            out_file = py_file.with_suffix(".out")
            if out_file.exists():
                name = f"{py_file.stem}"
                # Create a test item for the pair
                yield MypyPairItem.from_parent(
                    parent=self, name=name, py_path=py_file, out_path=out_file
                )


def pytest_collect_file(file_path, parent):
    """
    Hook that allows adding our MypyPairCollector when pytest collects files.
    We attach the collector to tests/conftest.py so the discovery runs once per tests session.
    """
    # Guard early so we don't attach the collector on older Pythons
    if version < (3, 10):
        return

    # Newer pytest passes pathlib.Path-like objects; ensure we can get a string/Path
    try:
        p = pathlib.Path(str(file_path))
    except Exception:
        p = pathlib.Path(getattr(file_path, "strpath", str(file_path)))

    # Only hook our collector on the tests/conftest.py file to avoid multiple runs
    if p.name == "conftest.py" and p.parent.name == "tests":
        if hasattr(MypyPairCollector, "from_parent"):
            return MypyPairCollector.from_parent(parent, path=file_path)
        else:
            # Fallback for very old pytest versions
            return MypyPairCollector(file_path, parent=parent)



import unittest

import wrapt


class Class:
    def __init__(self, value):
        self.value = value


class TestAttributeProxy(unittest.TestCase):

    def test_wrap_attribute(self):
        wrapt.wrap_object_attribute(__name__, "Class.value", wrapt.ObjectProxy)

        instance = Class(1)

        self.assertEqual(instance.value, 1)
        self.assertTrue(isinstance(instance.value, wrapt.ObjectProxy))

        instance.value = 2

        self.assertEqual(instance.value, 2)
        self.assertTrue(isinstance(instance.value, wrapt.ObjectProxy))



import unittest
import inspect
import types

import wrapt

DECORATORS_CODE = """
import wrapt

@wrapt.decorator
def passthru_decorator(wrapped, instance, args, kwargs):
    return wrapped(*args, **kwargs)
"""

decorators = types.ModuleType("decorators")
exec(DECORATORS_CODE, decorators.__dict__, decorators.__dict__)


def function1(arg):
    """documentation"""
    return arg


function1o = function1
function1d = decorators.passthru_decorator(function1)
assert function1d is not function1o


class TestNamingFunction(unittest.TestCase):

    def test_object_name(self):
        # Test preservation of function __name__ attribute.

        self.assertEqual(function1d.__name__, function1o.__name__)

    def test_object_qualname(self):
        # Test preservation of function __qualname__ attribute.

        try:
            __qualname__ = function1o.__qualname__
        except AttributeError:
            pass
        else:
            self.assertEqual(function1d.__qualname__, __qualname__)

    def test_module_name(self):
        # Test preservation of function __module__ attribute.

        self.assertEqual(function1d.__module__, __name__)

    def test_doc_string(self):
        # Test preservation of function __doc__ attribute.

        self.assertEqual(function1d.__doc__, function1o.__doc__)

    def test_argspec(self):
        # Test preservation of function argument specification.

        function1o_argspec = inspect.getfullargspec(function1o)
        function1d_argspec = inspect.getfullargspec(function1d)
        self.assertEqual(function1o_argspec, function1d_argspec)

    def test_getmembers(self):
        function1o_members = inspect.getmembers(function1o)
        function1d_members = inspect.getmembers(function1d)

    def test_isinstance(self):
        # Test preservation of isinstance() checks.

        self.assertTrue(isinstance(function1d, type(function1o)))


class TestCallingFunction(unittest.TestCase):

    def test_call_function(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        @wrapt.decorator
        def _decorator(wrapped, instance, args, kwargs):
            self.assertEqual(instance, None)
            self.assertEqual(args, _args)
            self.assertEqual(kwargs, _kwargs)
            return wrapped(*args, **kwargs)

        @_decorator
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))


if __name__ == "__main__":
    unittest.main()



import operator
import unittest

import wrapt


class TestDecorator(unittest.TestCase):

    def test_no_parameters(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        @wrapt.decorator
        def _decorator(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        @_decorator
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_instance_method_as_decorator(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Instance:
            def __init__(self):
                self.count = 0

            @wrapt.decorator
            def decorator(self, wrapped, instance, args, kwargs):
                self.count += 1
                return wrapped(*args, **kwargs)

        instance = Instance()

        @instance.decorator
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))
        self.assertEqual(instance.count, 1)

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))
        self.assertEqual(instance.count, 2)

    def test_class_method_as_decorator(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Instance:
            count = 0

            @wrapt.decorator
            @classmethod
            def decorator(cls, wrapped, instance, args, kwargs):
                cls.count += 1
                return wrapped(*args, **kwargs)

        @Instance.decorator
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))
        self.assertEqual(Instance.count, 1)

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))
        self.assertEqual(Instance.count, 2)

    def test_class_type_as_decorator(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        @wrapt.decorator
        class ClassDecorator:
            def __call__(self, wrapped, instance, args, kwargs):
                return wrapped(*args, **kwargs)

        @ClassDecorator
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_class_type_as_decorator_args(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        @wrapt.decorator
        class ClassDecorator:
            def __init__(self, arg):
                assert arg == 1

            def __call__(self, wrapped, instance, args, kwargs):
                return wrapped(*args, **kwargs)

        @ClassDecorator(arg=1)
        def _function(*args, **kwargs):
            return args, kwargs

        result = _function(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_decorated_function_as_class_attribute(self):

        @wrapt.decorator
        def passthrough(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        @passthrough
        def function(self):
            pass

        class A:
            _function = function

        self.assertTrue(A._function._self_parent is function)
        self.assertEqual(A._function._self_binding, "function")
        self.assertEqual(A._function._self_owner, A)

        a = A()

        A._function(a)

        self.assertTrue(a._function._self_parent is function)
        self.assertEqual(a._function._self_binding, "function")
        self.assertEqual(a._function._self_owner, A)

        a._function()

        # Test example without using the decorator to show same outcome.

        def xfunction(self):
            pass

        class B:
            _xfunction = xfunction

        b = B()

        B._xfunction(b)

        b._xfunction()

    def test_decorated_function_as_instance_attribute(self):

        @wrapt.decorator
        def passthrough(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        @passthrough
        def function(self):
            pass

        class A:
            def __init__(self):
                self._function = function

        a = A()

        self.assertTrue(a._function._self_parent is None)
        self.assertEqual(a._function._self_binding, "function")
        self.assertTrue(a._function._self_owner is None)

        a._function(a)

        bound_a = a._function.__get__(a, A)
        self.assertTrue(bound_a._self_parent is function)
        self.assertEqual(bound_a._self_binding, "function")
        self.assertTrue(bound_a._self_owner is A)

        bound_a()

        # Test example without using the decorator to show same outcome.

        def xfunction(self):
            pass

        class B:
            def __init__(self):
                self._xfunction = xfunction

        b = B()

        b._xfunction(b)

    def test_decorated_builtin_as_class_attribute(self):

        @wrapt.decorator
        def passthrough(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        function = passthrough(operator.add)

        class A:
            _function = function

        self.assertTrue(A._function._self_parent is None)
        self.assertEqual(A._function._self_binding, "builtin")
        self.assertTrue(A._function._self_owner is None)

        A._function(1, 2)

        a = A()

        self.assertTrue(a._function._self_parent is None)
        self.assertEqual(a._function._self_binding, "builtin")
        self.assertTrue(a._function._self_owner is None)

        a._function(1, 2)

        # Test example without using the decorator to show same outcome.

        class B:
            _xfunction = operator.add

        B._xfunction(1, 2)

        b = B()

        b._xfunction(1, 2)

    def test_call_semantics_for_assorted_decorator_use_cases(self):
        def g1():
            pass

        def g2(self):
            pass

        class C1:
            def __init__(self):
                self.f3 = g1

            def f1(self):
                print("SELF", self)

            f2 = g2

        c1 = C1()

        c1.f2()
        c1.f3()

        class C2:
            f2 = C1.f1
            f3 = C1.f2

        c2 = C2()

        c2.f2()
        c2.f3()

        class C3:
            f2 = c1.f1

        c3 = C3()

        c3.f2()

        @wrapt.decorator
        def passthrough(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class C11:
            @passthrough
            def f1(self):
                print("SELF", self)

        c11 = C11()

        class C12:
            f2 = C11.f1

        c12 = C12()

        c12.f2()

        class C13:
            f2 = c11.f1

        c13 = C13()

        c13.f2()

        C11.f1(c11)
        C12.f2(c12)

    def test_call_semantics_for_assorted_wrapped_descriptor_use_cases(self):
        class A:
            def __call__(self):
                print("A:__call__")

        a = A()

        class B:
            def __call__(self):
                print("B:__call__")

            def __get__(self, obj, type):
                print("B:__get__")
                return self

        b = B()

        class C:
            f1 = a
            f2 = b

        c = C()

        c.f1()
        c.f2()

        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class D:
            f1 = wrapper(a)
            f2 = wrapper(b)

        d = D()

        d.f1()
        d.f2()


if __name__ == "__main__":
    unittest.main()



"""This module implements decorators for implementing other decorators
as well as some commonly used decorators.

"""

import sys
from functools import partial
from inspect import isclass, signature
from threading import Lock, RLock

from .__wrapt__ import BoundFunctionWrapper, CallableObjectProxy, FunctionWrapper
from .arguments import formatargspec

# Adapter wrapper for the wrapped function which will overlay certain
# properties from the adapter function onto the wrapped function so that
# functions such as inspect.getfullargspec(), inspect.signature() and
# inspect.getsource() return the correct results one would expect.


class _AdapterFunctionCode(CallableObjectProxy):

    def __init__(self, wrapped_code, adapter_code):
        super(_AdapterFunctionCode, self).__init__(wrapped_code)
        self._self_adapter_code = adapter_code

    @property
    def co_argcount(self):
        return self._self_adapter_code.co_argcount

    @property
    def co_code(self):
        return self._self_adapter_code.co_code

    @property
    def co_flags(self):
        return self._self_adapter_code.co_flags

    @property
    def co_kwonlyargcount(self):
        return self._self_adapter_code.co_kwonlyargcount

    @property
    def co_varnames(self):
        return self._self_adapter_code.co_varnames


class _AdapterFunctionSurrogate(CallableObjectProxy):

    def __init__(self, wrapped, adapter):
        super(_AdapterFunctionSurrogate, self).__init__(wrapped)
        self._self_adapter = adapter

    @property
    def __code__(self):
        return _AdapterFunctionCode(
            self.__wrapped__.__code__, self._self_adapter.__code__
        )

    @property
    def __defaults__(self):
        return self._self_adapter.__defaults__

    @property
    def __kwdefaults__(self):
        return self._self_adapter.__kwdefaults__

    @property
    def __signature__(self):
        if "signature" not in globals():
            return self._self_adapter.__signature__
        else:
            return signature(self._self_adapter)


class _BoundAdapterWrapper(BoundFunctionWrapper):

    @property
    def __func__(self):
        return _AdapterFunctionSurrogate(
            self.__wrapped__.__func__, self._self_parent._self_adapter
        )

    @property
    def __signature__(self):
        if "signature" not in globals():
            return self.__wrapped__.__signature__
        else:
            return signature(self._self_parent._self_adapter)


class AdapterWrapper(FunctionWrapper):

    __bound_function_wrapper__ = _BoundAdapterWrapper

    def __init__(self, *args, **kwargs):
        adapter = kwargs.pop("adapter")
        super(AdapterWrapper, self).__init__(*args, **kwargs)
        self._self_surrogate = _AdapterFunctionSurrogate(self.__wrapped__, adapter)
        self._self_adapter = adapter

    @property
    def __code__(self):
        return self._self_surrogate.__code__

    @property
    def __defaults__(self):
        return self._self_surrogate.__defaults__

    @property
    def __kwdefaults__(self):
        return self._self_surrogate.__kwdefaults__

    @property
    def __signature__(self):
        return self._self_surrogate.__signature__


class AdapterFactory:
    def __call__(self, wrapped):
        raise NotImplementedError()


class DelegatedAdapterFactory(AdapterFactory):
    def __init__(self, factory):
        super(DelegatedAdapterFactory, self).__init__()
        self.factory = factory

    def __call__(self, wrapped):
        return self.factory(wrapped)


adapter_factory = DelegatedAdapterFactory

# Decorator for creating other decorators. This decorator and the
# wrappers which they use are designed to properly preserve any name
# attributes, function signatures etc, in addition to the wrappers
# themselves acting like a transparent proxy for the original wrapped
# function so the wrapper is effectively indistinguishable from the
# original wrapped function.


def decorator(wrapper=None, /, *, enabled=None, adapter=None, proxy=FunctionWrapper):
    """
    The decorator should be supplied with a single positional argument
    which is the `wrapper` function to be used to implement the
    decorator. This may be preceded by a step whereby the keyword
    arguments are supplied to customise the behaviour of the
    decorator. The `adapter` argument is used to optionally denote a
    separate function which is notionally used by an adapter
    decorator. In that case parts of the function `__code__` and
    `__defaults__` attributes are used from the adapter function
    rather than those of the wrapped function. This allows for the
    argument specification from `inspect.getfullargspec()` and similar
    functions to be overridden with a prototype for a different
    function than what was wrapped. The `enabled` argument provides a
    way to enable/disable the use of the decorator. If the type of
    `enabled` is a boolean, then it is evaluated immediately and the
    wrapper not even applied if it is `False`. If not a boolean, it will
    be evaluated when the wrapper is called for an unbound wrapper,
    and when binding occurs for a bound wrapper. When being evaluated,
    if `enabled` is callable it will be called to obtain the value to
    be checked. If `False`, the wrapper will not be called and instead
    the original wrapped function will be called directly instead.
    The `proxy` argument provides a way of passing a custom version of
    the `FunctionWrapper` class used in decorating the function.
    """

    if wrapper is not None:
        # Helper function for creating wrapper of the appropriate
        # time when we need it down below.

        def _build(wrapped, wrapper, enabled=None, adapter=None):
            if adapter:
                if isinstance(adapter, AdapterFactory):
                    adapter = adapter(wrapped)

                if not callable(adapter):
                    ns = {}

                    # Check if the signature argument specification has
                    # annotations. If it does then we need to remember
                    # it but also drop it when attempting to manufacture
                    # a standin adapter function. This is necessary else
                    # it will try and look up any types referenced in
                    # the annotations in the empty namespace we use,
                    # which will fail.

                    annotations = {}

                    if not isinstance(adapter, str):
                        if len(adapter) == 7:
                            annotations = adapter[-1]
                            adapter = adapter[:-1]
                        adapter = formatargspec(*adapter)

                    exec(f"def adapter{adapter}: pass", ns, ns)
                    adapter = ns["adapter"]

                    # Override the annotations for the manufactured
                    # adapter function so they match the original
                    # adapter signature argument specification.

                    if annotations:
                        adapter.__annotations__ = annotations

                return AdapterWrapper(
                    wrapped=wrapped, wrapper=wrapper, enabled=enabled, adapter=adapter
                )

            return proxy(wrapped=wrapped, wrapper=wrapper, enabled=enabled)

        # The wrapper has been provided so return the final decorator.
        # The decorator is itself one of our function wrappers so we
        # can determine when it is applied to functions, instance methods
        # or class methods. This allows us to bind the instance or class
        # method so the appropriate self or cls attribute is supplied
        # when it is finally called.

        def _wrapper(wrapped, instance, args, kwargs):
            # We first check for the case where the decorator was applied
            # to a class type.
            #
            #     @decorator
            #     class mydecoratorclass:
            #         def __init__(self, arg=None):
            #             self.arg = arg
            #         def __call__(self, wrapped, instance, args, kwargs):
            #             return wrapped(*args, **kwargs)
            #
            #     @mydecoratorclass(arg=1)
            #     def function():
            #         pass
            #
            # In this case an instance of the class is to be used as the
            # decorator wrapper function. If args was empty at this point,
            # then it means that there were optional keyword arguments
            # supplied to be used when creating an instance of the class
            # to be used as the wrapper function.

            if instance is None and isclass(wrapped) and not args:
                # We still need to be passed the target function to be
                # wrapped as yet, so we need to return a further function
                # to be able to capture it.

                def _capture(target_wrapped):
                    # Now have the target function to be wrapped and need
                    # to create an instance of the class which is to act
                    # as the decorator wrapper function. Before we do that,
                    # we need to first check that use of the decorator
                    # hadn't been disabled by a simple boolean. If it was,
                    # the target function to be wrapped is returned instead.

                    _enabled = enabled
                    if type(_enabled) is bool:
                        if not _enabled:
                            return target_wrapped
                        _enabled = None

                    # Now create an instance of the class which is to act
                    # as the decorator wrapper function. Any arguments had
                    # to be supplied as keyword only arguments so that is
                    # all we pass when creating it.

                    target_wrapper = wrapped(**kwargs)

                    # Finally build the wrapper itself and return it.

                    return _build(target_wrapped, target_wrapper, _enabled, adapter)

                return _capture

            # We should always have the target function to be wrapped at
            # this point as the first (and only) value in args.

            target_wrapped = args[0]

            # Need to now check that use of the decorator hadn't been
            # disabled by a simple boolean. If it was, then target
            # function to be wrapped is returned instead.

            _enabled = enabled
            if type(_enabled) is bool:
                if not _enabled:
                    return target_wrapped
                _enabled = None

            # We now need to build the wrapper, but there are a couple of
            # different cases we need to consider.

            if instance is None:
                if isclass(wrapped):
                    # In this case the decorator was applied to a class
                    # type but optional keyword arguments were not supplied
                    # for initialising an instance of the class to be used
                    # as the decorator wrapper function.
                    #
                    #     @decorator
                    #     class mydecoratorclass:
                    #         def __init__(self, arg=None):
                    #             self.arg = arg
                    #         def __call__(self, wrapped, instance,
                    #                 args, kwargs):
                    #             return wrapped(*args, **kwargs)
                    #
                    #     @mydecoratorclass
                    #     def function():
                    #         pass
                    #
                    # We still need to create an instance of the class to
                    # be used as the decorator wrapper function, but no
                    # arguments are pass.

                    target_wrapper = wrapped()

                else:
                    # In this case the decorator was applied to a normal
                    # function, or possibly a static method of a class.
                    #
                    #     @decorator
                    #     def mydecoratorfuntion(wrapped, instance,
                    #             args, kwargs):
                    #         return wrapped(*args, **kwargs)
                    #
                    #     @mydecoratorfunction
                    #     def function():
                    #         pass
                    #
                    # That normal function becomes the decorator wrapper
                    # function.

                    target_wrapper = wrapper

            else:
                if isclass(instance):
                    # In this case the decorator was applied to a class
                    # method.
                    #
                    #     class myclass:
                    #         @decorator
                    #         @classmethod
                    #         def decoratorclassmethod(cls, wrapped,
                    #                 instance, args, kwargs):
                    #             return wrapped(*args, **kwargs)
                    #
                    #     instance = myclass()
                    #
                    #     @instance.decoratorclassmethod
                    #     def function():
                    #         pass
                    #
                    # This one is a bit strange because binding was actually
                    # performed on the wrapper created by our decorator
                    # factory. We need to apply that binding to the decorator
                    # wrapper function that the decorator factory
                    # was applied to.

                    target_wrapper = wrapper.__get__(None, instance)

                else:
                    # In this case the decorator was applied to an instance
                    # method.
                    #
                    #     class myclass:
                    #         @decorator
                    #         def decoratorclassmethod(self, wrapped,
                    #                 instance, args, kwargs):
                    #             return wrapped(*args, **kwargs)
                    #
                    #     instance = myclass()
                    #
                    #     @instance.decoratorclassmethod
                    #     def function():
                    #         pass
                    #
                    # This one is a bit strange because binding was actually
                    # performed on the wrapper created by our decorator
                    # factory. We need to apply that binding to the decorator
                    # wrapper function that the decorator factory
                    # was applied to.

                    target_wrapper = wrapper.__get__(instance, type(instance))

            # Finally build the wrapper itself and return it.

            return _build(target_wrapped, target_wrapper, _enabled, adapter)

        # We first return our magic function wrapper here so we can
        # determine in what context the decorator factory was used. In
        # other words, it is itself a universal decorator. The decorator
        # function is used as the adapter so that linters see a signature
        # corresponding to the decorator and not the wrapper it is being
        # applied to.

        return _build(wrapper, _wrapper, adapter=decorator)

    else:
        # The wrapper still has not been provided, so we are just
        # collecting the optional keyword arguments. Return the
        # decorator again wrapped in a partial using the collected
        # arguments.

        return partial(decorator, enabled=enabled, adapter=adapter, proxy=proxy)


# Decorator for implementing thread synchronization. It can be used as a
# decorator, in which case the synchronization context is determined by
# what type of function is wrapped, or it can also be used as a context
# manager, where the user needs to supply the correct synchronization
# context. It is also possible to supply an object which appears to be a
# synchronization primitive of some sort, by virtue of having release()
# and acquire() methods. In that case that will be used directly as the
# synchronization primitive without creating a separate lock against the
# derived or supplied context.


def synchronized(wrapped):
    """Depending on the nature of the `wrapped` object, will either return a
    decorator which can be used to wrap a function or method, or a context
    manager, both of which will act accordingly depending on how used, to
    synchronize access to calling of the wrapped function, or the block of
    code within the context manager. If it is an object which is a
    synchronization primitive, such as a threading Lock, RLock, Semaphore,
    Condition, or Event, then it is assumed that the object is to be used
    directly as the synchronization primitive, otherwise a lock is created
    automatically and attached to the wrapped object and used as the
    synchronization primitive.
    """

    # Determine if being passed an object which is a synchronization
    # primitive. We can't check by type for Lock, RLock, Semaphore etc,
    # as the means of creating them isn't the type. Therefore use the
    # existence of acquire() and release() methods. This is more
    # extensible anyway as it allows custom synchronization mechanisms.

    if hasattr(wrapped, "acquire") and hasattr(wrapped, "release"):
        # We remember what the original lock is and then return a new
        # decorator which accesses and locks it. When returning the new
        # decorator we wrap it with an object proxy so we can override
        # the context manager methods in case it is being used to wrap
        # synchronized statements with a 'with' statement.

        lock = wrapped

        @decorator
        def _synchronized(wrapped, instance, args, kwargs):
            # Execute the wrapped function while the original supplied
            # lock is held.

            with lock:
                return wrapped(*args, **kwargs)

        class _PartialDecorator(CallableObjectProxy):

            def __enter__(self):
                lock.acquire()
                return lock

            def __exit__(self, *args):
                lock.release()

        return _PartialDecorator(wrapped=_synchronized)

    # Following only apply when the lock is being created automatically
    # based on the context of what was supplied. In this case we supply
    # a final decorator, but need to use FunctionWrapper directly as we
    # want to derive from it to add context manager methods in case it is
    # being used to wrap synchronized statements with a 'with' statement.

    def _synchronized_lock(context):
        # Attempt to retrieve the lock for the specific context.

        lock = vars(context).get("_synchronized_lock", None)

        if lock is None:
            # There is no existing lock defined for the context we
            # are dealing with so we need to create one. This needs
            # to be done in a way to guarantee there is only one
            # created, even if multiple threads try and create it at
            # the same time. We can't always use the setdefault()
            # method on the __dict__ for the context. This is the
            # case where the context is a class, as __dict__ is
            # actually a dictproxy. What we therefore do is use a
            # meta lock on this wrapper itself, to control the
            # creation and assignment of the lock attribute against
            # the context.

            with synchronized._synchronized_meta_lock:
                # We need to check again for whether the lock we want
                # exists in case two threads were trying to create it
                # at the same time and were competing to create the
                # meta lock.

                lock = vars(context).get("_synchronized_lock", None)

                if lock is None:
                    lock = RLock()
                    setattr(context, "_synchronized_lock", lock)

        return lock

    def _synchronized_wrapper(wrapped, instance, args, kwargs):
        # Execute the wrapped function while the lock for the
        # desired context is held. If instance is None then the
        # wrapped function is used as the context.

        with _synchronized_lock(instance if instance is not None else wrapped):
            return wrapped(*args, **kwargs)

    class _FinalDecorator(FunctionWrapper):

        def __enter__(self):
            self._self_lock = _synchronized_lock(self.__wrapped__)
            self._self_lock.acquire()
            return self._self_lock

        def __exit__(self, *args):
            self._self_lock.release()

    return _FinalDecorator(wrapped=wrapped, wrapper=_synchronized_wrapper)


synchronized._synchronized_meta_lock = Lock()  # type: ignore[attr-defined]



"""
This example demonstrates the usage of the patch_wrapper_function() function.
"""

import sys
from typing import Any, Callable

from wrapt import patch_function_wrapper


def function(x: int, y: str = "default") -> str:
    """A simple function to be wrapped."""
    return f"{x}: {y}"


this_module = sys.modules[__name__]


@patch_function_wrapper(this_module, "function")
def patch_module_direct(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


@patch_function_wrapper(this_module, "function")
def patch_module_using_name(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


# Incorrect type for attribute path. (FAIL)
@patch_function_wrapper(this_module, None)
def incorrect_type_for_attribute_path(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


# Incorrect prototype for wrapper function. (FAIL)
@patch_function_wrapper(this_module, "function")
def incorrect_prototype_for_wrapper(
    wrapped: Callable[..., Any], args: tuple[Any, ...], kwargs: dict[str, Any]
) -> Any:
    return wrapped(*args, **kwargs)


@patch_function_wrapper(this_module, "function", enabled=True)
def boolean_for_enabled_argument(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


@patch_function_wrapper(this_module, "function", enabled=lambda: True)
def callable_for_enabled_argument(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


@patch_function_wrapper(this_module, "function", enabled=None)
def none_for_enabled_argument(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)


# Incorrect type for enabled argument. (FAIL)
@patch_function_wrapper(this_module, "function", enabled="string")
def incorrect_type_for_enabled_argument(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    return wrapped(*args, **kwargs)



"""
This example demonstrates the usage of the FunctionWrapper type.

It covers the following cases:
- Default to enabled. (OKAY)
- Passing None to enabled. (OKAY)
- Passing bool to enabled. (OKAY)
- Passing a callable to enabled. (OKAY)
- Passing incorrect types to enabled. (FAIL)
"""

from typing import Any, Callable

from wrapt import FunctionWrapper


def function(x: int, y: str = "default") -> str:
    """A simple function to be wrapped."""
    return f"{x}: {y}"


def standard_wrapper(
    wrapped: Callable[..., Any],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Any:
    try:
        print(f"Before calling {wrapped.__name__}")
        return wrapped(*args, **kwargs)
    finally:
        print(f"After calling {wrapped.__name__}")


enabled_default = FunctionWrapper(function, standard_wrapper)
enabled_none = FunctionWrapper(function, standard_wrapper, None)
enabled_true = FunctionWrapper(function, standard_wrapper, True)
enabled_false = FunctionWrapper(function, standard_wrapper, False)
enabled_callable = FunctionWrapper(function, standard_wrapper, lambda: True)

enabled_string = FunctionWrapper(function, standard_wrapper, "string")



import sys
import unittest

import wrapt


class TestLazyObjectProxy(unittest.TestCase):

    def test_lazy_object_proxy(self):
        status = {"created": 0}

        def factory():
            status["created"] += 1
            return 42

        proxy = wrapt.LazyObjectProxy(factory)

        self.assertEqual(status["created"], 0)
        self.assertEqual(proxy.__wrapped__, 42)
        self.assertEqual(int(proxy), 42)
        self.assertEqual(status["created"], 1)
        self.assertEqual(int(proxy), 42)
        self.assertEqual(status["created"], 1)

    def test_lazy_import(self):
        if "sched" in sys.modules:
            del sys.modules["sched"]

        sched_mod = wrapt.lazy_import("sched")

        self.assertFalse("sched" in sys.modules)

        self.assertIsNotNone(sched_mod.scheduler)

        self.assertTrue("sched" in sys.modules)

    def test_lazy_import_attribute(self):
        if "sched" in sys.modules:
            del sys.modules["sched"]

        scheduler = wrapt.lazy_import("sched", "scheduler")

        self.assertFalse("sched" in sys.modules)

        import sched

        self.assertIsNotNone(scheduler.__wrapped__)
        self.assertIs(scheduler.__wrapped__, sched.scheduler)

        self.assertTrue("sched" in sys.modules)

    def test_lazy_import_dotted(self):
        if "http.client" in sys.modules:
            del sys.modules["http.client"]

        client_mod = wrapt.lazy_import("http.client")

        self.assertFalse("http.client" in sys.modules)

        self.assertIsNotNone(client_mod.HTTPConnection)

        self.assertTrue("http.client" in sys.modules)

    def test_lazy_import_callable(self):
        dumps = wrapt.lazy_import("json", "dumps")

        self.assertTrue(callable(dumps))

        result = dumps({"key": "value"})

        self.assertEqual(result, '{"key": "value"}')

    def test_lazy_import_iterable(self):
        from collections.abc import Iterable

        if "string" in sys.modules:
            del sys.modules["string"]

        digits0 = wrapt.lazy_import("string", "digits")

        self.assertFalse(hasattr(type(digits0), "__iter__"))

        self.assertTrue(callable(digits0))

        # XXX Iteration does not seem to test for __iter__ being a method
        # on the type, so this test is not sufficient to raise the TypeError.
        # I was at some point seeing expected failures here, but not now and
        # I don't know why.

        # if "string" in sys.modules:
        #     del sys.modules["string"]

        # digits1 = wrapt.lazy_import("string", "digits")

        # with self.assertRaises(TypeError):
        #     for char in digits1:
        #         pass

        if "string" in sys.modules:
            del sys.modules["string"]

        digits2 = wrapt.lazy_import("string", "digits", interface=Iterable)

        self.assertTrue(hasattr(type(digits2), "__iter__"))

        self.assertFalse(callable(digits2))

        for char in digits2:
            self.assertIn(char, "0123456789")

        self.assertEqual("".join(digits2), "0123456789")



import unittest

import pickle

import wrapt


class CustomObjectProxy(wrapt.ObjectProxy):

    def __reduce_ex__(self, proto):
        return (list, (self.__wrapped__,))


class TestObjectPickle(unittest.TestCase):

    def test_pickle(self):
        proxy = wrapt.ObjectProxy([1])

        with self.assertRaises(NotImplementedError) as context:
            data = pickle.dumps(proxy)

        self.assertTrue(
            str(context.exception) == "object proxy must define __reduce_ex__()"
        )

    def test_pickle_proxy(self):
        proxy1 = CustomObjectProxy([1])
        pickled = pickle.dumps(proxy1)
        restored = pickle.loads(pickled)

        self.assertEqual(proxy1.__wrapped__, restored)


if __name__ == "__main__":
    unittest.main()



import re
import sys
import types
import unittest

is_pypy = "__pypy__" in sys.builtin_module_names

import wrapt

OBJECTS_CODE = """
class TargetBaseClass:
    "documentation"

class Target(TargetBaseClass):
    "documentation"

def target():
    "documentation"
    pass
"""

objects = types.ModuleType("objects")
exec(OBJECTS_CODE, objects.__dict__, objects.__dict__)


class TestAttributeAccess(unittest.TestCase):

    def test_init_not_called(self):
        a = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        b = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            a.__wrapped__
        except ValueError:
            pass

        try:
            a + b
        except ValueError:
            pass

        try:
            a.__wrapped__
        except AttributeError:
            pass

        try:
            a + b
        except AttributeError:
            pass

    def test_attributes(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        self.assertEqual(function2.__wrapped__, function1)

    def test_get_wrapped(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        self.assertEqual(function2.__wrapped__, function1)

        function3 = wrapt.ObjectProxy(function2)

        self.assertEqual(function3.__wrapped__, function1)

    def test_set_wrapped(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        self.assertEqual(function2, function1)
        self.assertEqual(function2.__wrapped__, function1)
        self.assertEqual(function2.__name__, function1.__name__)

        self.assertEqual(function2.__qualname__, function1.__qualname__)

        function2.__wrapped__ = None

        self.assertFalse(hasattr(function1, "__wrapped__"))

        self.assertEqual(function2, None)
        self.assertEqual(function2.__wrapped__, None)
        self.assertFalse(hasattr(function2, "__name__"))

        self.assertFalse(hasattr(function2, "__qualname__"))

        def function3(*args, **kwargs):
            return args, kwargs

        function2.__wrapped__ = function3

        self.assertEqual(function2, function3)
        self.assertEqual(function2.__wrapped__, function3)
        self.assertEqual(function2.__name__, function3.__name__)

        self.assertEqual(function2.__qualname__, function3.__qualname__)

    def test_delete_wrapped(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        def run(*args):
            del function2.__wrapped__

        self.assertRaises(TypeError, run, ())

    def test_proxy_attribute(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        function2._self_variable = True

        self.assertFalse(hasattr(function1, "_self_variable"))
        self.assertTrue(hasattr(function2, "_self_variable"))

        self.assertEqual(function2._self_variable, True)

        del function2._self_variable

        self.assertFalse(hasattr(function1, "_self_variable"))
        self.assertFalse(hasattr(function2, "_self_variable"))

        self.assertEqual(getattr(function2, "_self_variable", None), None)

    def test_wrapped_attribute(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        function2.variable = True

        self.assertTrue(hasattr(function1, "variable"))
        self.assertTrue(hasattr(function2, "variable"))

        self.assertEqual(function2.variable, True)

        del function2.variable

        self.assertFalse(hasattr(function1, "variable"))
        self.assertFalse(hasattr(function2, "variable"))

        self.assertEqual(getattr(function2, "variable", None), None)

    def test_attribute_lookup_modified(self):
        class Object:
            @property
            def value(self):
                return "value"

        class WrappedObject(wrapt.ObjectProxy):
            @property
            def value(self):
                return 2 * self.__wrapped__.value

        WrappedObject(Object()).value == "valuevalue"

    def test_attribute_lookup_value_exception(self):
        class Object:
            @property
            def value(self):
                return "value"

        class WrappedObject(wrapt.ObjectProxy):
            @property
            def value(self):
                raise ValueError("value-error")

        try:
            WrappedObject(Object()).value == "value"

        except ValueError as e:
            pass

        else:
            raise RuntimeError("should not fail here")

    def test_attribute_lookup_attribute_exception(self):
        class Object:
            @property
            def value(self):
                return "value"

        class WrappedObject(wrapt.ObjectProxy):
            @property
            def value(self):
                raise AttributeError("attribute-error")

        # Raising of an AttributeError in this case is a sort of odd situation
        # because the exception results in it being determined there was no
        # wrapper for the value attribute and so it returns the original value
        # instead and robs the wrapper of the chance to return an alternate
        # value.

        WrappedObject(Object()).value == "value"


class TestNamingObjectProxy(unittest.TestCase):

    def test_class_object_name(self):
        # Test preservation of class __name__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__name__, target.__name__)

    def test_class_object_qualname(self):
        # Test preservation of class __qualname__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        try:
            __qualname__ = target.__qualname__
        except AttributeError:
            pass
        else:
            self.assertEqual(wrapper.__qualname__, __qualname__)

    def test_class_module_name(self):
        # Test preservation of class __module__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__module__, target.__module__)

    def test_class_doc_string(self):
        # Test preservation of class __doc__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__doc__, target.__doc__)

    def test_instance_module_name(self):
        # Test preservation of instance __module__ attribute.

        target = objects.Target()
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__module__, target.__module__)

    def test_instance_doc_string(self):
        # Test preservation of instance __doc__ attribute.

        target = objects.Target()
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__doc__, target.__doc__)

    def test_function_object_name(self):
        # Test preservation of function __name__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__name__, target.__name__)

    def test_function_object_qualname(self):
        # Test preservation of function __qualname__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        try:
            __qualname__ = target.__qualname__
        except AttributeError:
            pass
        else:
            self.assertEqual(wrapper.__qualname__, __qualname__)

    def test_function_module_name(self):
        # Test preservation of function __module__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__module__, target.__module__)

    def test_function_doc_string(self):
        # Test preservation of function __doc__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__doc__, target.__doc__)


class TestTypeObjectProxy(unittest.TestCase):

    def test_class_of_class(self):
        # Test preservation of class __class__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__class__, target.__class__)

        self.assertTrue(isinstance(wrapper, type(target)))

    def test_class_of_instance(self):
        # Test preservation of instance __class__ attribute.

        target = objects.Target()
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__class__, target.__class__)

        self.assertTrue(isinstance(wrapper, objects.Target))
        self.assertTrue(isinstance(wrapper, objects.TargetBaseClass))

    def test_class_of_function(self):
        # Test preservation of function __class__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(wrapper.__class__, target.__class__)

        self.assertTrue(isinstance(wrapper, type(target)))


class TestDirObjectProxy(unittest.TestCase):

    def test_dir_of_class(self):
        # Test preservation of class __dir__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(dir(wrapper), dir(target))

    def test_vars_of_class(self):
        # Test preservation of class __dir__ attribute.

        target = objects.Target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(vars(wrapper), vars(target))

    def test_dir_of_instance(self):
        # Test preservation of instance __dir__ attribute.

        target = objects.Target()
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(dir(wrapper), dir(target))

    def test_vars_of_instance(self):
        # Test preservation of instance __dir__ attribute.

        target = objects.Target()
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(vars(wrapper), vars(target))

    def test_dir_of_function(self):
        # Test preservation of function __dir__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(dir(wrapper), dir(target))

    def test_vars_of_function(self):
        # Test preservation of function __dir__ attribute.

        target = objects.target
        wrapper = wrapt.ObjectProxy(target)

        self.assertEqual(vars(wrapper), vars(target))


class TestCallingObject(unittest.TestCase):

    def test_function_no_args(self):
        _args = ()
        _kwargs = {}

        def function(*args, **kwargs):
            return args, kwargs

        wrapper = wrapt.CallableObjectProxy(function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_function_args(self):
        _args = (1, 2)
        _kwargs = {}

        def function(*args, **kwargs):
            return args, kwargs

        wrapper = wrapt.CallableObjectProxy(function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_function_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        def function(*args, **kwargs):
            return args, kwargs

        wrapper = wrapt.CallableObjectProxy(function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_function_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        def function(*args, **kwargs):
            return args, kwargs

        wrapper = wrapt.CallableObjectProxy(function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_via_class_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(Class())

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_via_class_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(Class(), *_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_via_class_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(Class(), **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_instancemethod_via_class_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            def function(self, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(Class(), *_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_via_class_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_via_class_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_via_class_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_classmethod_via_class_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @classmethod
            def function(cls, *args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class().function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_via_class_no_args(self):
        _args = ()
        _kwargs = {}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_via_class_args(self):
        _args = (1, 2)
        _kwargs = {}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(*_args)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_via_class_kwargs(self):
        _args = ()
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(**_kwargs)

        self.assertEqual(result, (_args, _kwargs))

    def test_staticmethod_via_class_args_plus_kwargs(self):
        _args = (1, 2)
        _kwargs = {"one": 1, "two": 2}

        class Class:
            @staticmethod
            def function(*args, **kwargs):
                return args, kwargs

        wrapper = wrapt.CallableObjectProxy(Class.function)

        result = wrapper(*_args, **_kwargs)

        self.assertEqual(result, (_args, _kwargs))


class TestCallingObjectAuto(unittest.TestCase):

    def test_function_no_args(self):
        _args = ()
        _kwargs = {}

        def function(*args, **kwargs):
            return args, kwargs

        wrapper = wrapt.AutoObjectProxy(function)

        self.assertTrue(callable(wrapper))

        result = wrapper()

        self.assertEqual(result, (_args, _kwargs))


class TestIterObjectProxy(unittest.TestCase):

    def test_iteration(self):
        items = [1, 2]

        wrapper = wrapt.ObjectProxy(items)

        result = [x for x in wrapper]

        self.assertEqual(result, items)


class TestContextManagerObjectProxy(unittest.TestCase):

    def test_context_manager(self):
        class Class:
            def __enter__(self):
                return self

            def __exit__(*args, **kwargs):
                return

        instance = Class()

        wrapper = wrapt.ObjectProxy(instance)

        with wrapper:
            pass

    def test_async_context_manager(self):
        class Class:
            async def __aenter__(self):
                return self

            async def __aexit__(*args, **kwargs):
                return

        instance = Class()

        wrapper = wrapt.ObjectProxy(instance)

        async def run():
            async with wrapper:
                pass

        import asyncio

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(run())
        loop.close()


class TestEqualityObjectProxy(unittest.TestCase):

    def test_object_hash(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        self.assertEqual(hash(function2), hash(function1))

    def test_mapping_key(self):
        def function1(*args, **kwargs):
            return args, kwargs

        function2 = wrapt.ObjectProxy(function1)

        table = dict()
        table[function1] = True

        self.assertTrue(table.get(function2))

        table = dict()
        table[function2] = True

        self.assertTrue(table.get(function1))

    def test_comparison(self):
        one = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertTrue(two > 1)
        self.assertTrue(two >= 1)
        self.assertTrue(two < 3)
        self.assertTrue(two <= 3)
        self.assertTrue(two != 1)
        self.assertTrue(two == 2)
        self.assertTrue(two != 3)

        self.assertTrue(2 > one)
        self.assertTrue(2 >= one)
        self.assertTrue(2 < three)
        self.assertTrue(2 <= three)
        self.assertTrue(2 != one)
        self.assertTrue(2 == two)
        self.assertTrue(2 != three)

        self.assertTrue(two > one)
        self.assertTrue(two >= one)
        self.assertTrue(two < three)
        self.assertTrue(two <= three)
        self.assertTrue(two != one)
        self.assertTrue(two == two)
        self.assertTrue(two != three)


class TestAsNumberObjectProxy(unittest.TestCase):

    def test_nonzero(self):
        true = wrapt.ObjectProxy(True)
        false = wrapt.ObjectProxy(False)

        self.assertTrue(true)
        self.assertFalse(false)

        self.assertTrue(bool(true))
        self.assertFalse(bool(false))

        self.assertTrue(not false)
        self.assertFalse(not true)

    def test_int(self):
        one = wrapt.ObjectProxy(1)

        self.assertEqual(int(one), 1)

    def test_float(self):
        one = wrapt.ObjectProxy(1)

        self.assertEqual(float(one), 1.0)

    def test_add(self):
        one = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)

        self.assertEqual(one + two, 1 + 2)
        self.assertEqual(1 + two, 1 + 2)
        self.assertEqual(one + 2, 1 + 2)

    def test_add_uninitialized_args(self):
        result = object()

        one = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        two = wrapt.ObjectProxy(2)

        try:
            assert one + two == result
        except ValueError:
            pass

        one = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert one + two == result
        except ValueError:
            pass

    def test_sub(self):
        one = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)

        self.assertEqual(one - two, 1 - 2)
        self.assertEqual(1 - two, 1 - 2)
        self.assertEqual(one - 2, 1 - 2)

    def test_sub_uninitialized_args(self):
        result = object()

        one = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        two = wrapt.ObjectProxy(2)

        try:
            assert one - two == result
        except ValueError:
            pass

        one = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert one - two == result
        except ValueError:
            pass

    def test_mul(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(two * three, 2 * 3)
        self.assertEqual(2 * three, 2 * 3)
        self.assertEqual(two * 3, 2 * 3)

    def test_mul_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert two * three == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert two * three == result
        except ValueError:
            pass

    def test_div(self):
        # On Python 2 this will pick up div and on Python
        # 3 it will pick up truediv.

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(two / three, 2 / 3)
        self.assertEqual(2 / three, 2 / 3)
        self.assertEqual(two / 3, 2 / 3)

    def test_div_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert two / three == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert two / three == result
        except ValueError:
            pass

    def test_floordiv(self):
        two = wrapt.ObjectProxy(2)
        four = wrapt.ObjectProxy(4)

        self.assertEqual(four // two, 4 // 2)
        self.assertEqual(4 // two, 4 // 2)
        self.assertEqual(four // 2, 4 // 2)

    def test_floordiv_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        four = wrapt.ObjectProxy(4)

        try:
            assert two // four == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        four = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert two // four == result
        except ValueError:
            pass

    def test_mod(self):
        two = wrapt.ObjectProxy(2)
        four = wrapt.ObjectProxy(4)

        self.assertEqual(four % two, 4 % 2)
        self.assertEqual(4 % two, 4 % 2)
        self.assertEqual(four % 2, 4 % 2)

    def test_mod_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        four = wrapt.ObjectProxy(4)

        try:
            assert two % four == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        four = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert two % four == result
        except ValueError:
            pass

    def test_divmod(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(divmod(three, two), divmod(3, 2))
        self.assertEqual(divmod(3, two), divmod(3, 2))
        self.assertEqual(divmod(three, 2), divmod(3, 2))

    def test_divmod_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert divmod(two, three) == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert divmod(two, three) == result
        except ValueError:
            pass

    def test_pow(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three**two, pow(3, 2))
        self.assertEqual(3**two, pow(3, 2))
        self.assertEqual(three**2, pow(3, 2))

        self.assertEqual(pow(three, two), pow(3, 2))
        self.assertEqual(pow(3, two), pow(3, 2))
        self.assertEqual(pow(three, 2), pow(3, 2))

        # Only PyPy implements __rpow__ for ternary pow().
        # Note: No longer test for this as pypy 3.9+ seems
        # to have been updated to not support ternary pow()
        # in same way.

        # if is_pypy:
        #     self.assertEqual(pow(three, two, 2), pow(3, 2, 2))
        #     self.assertEqual(pow(3, two, 2), pow(3, 2, 2))

        self.assertEqual(pow(three, 2, 2), pow(3, 2, 2))

    def test_pow_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three**two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three**two == result
        except ValueError:
            pass

    def test_lshift(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three << two, 3 << 2)
        self.assertEqual(3 << two, 3 << 2)
        self.assertEqual(three << 2, 3 << 2)

    def test_lshift_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three << two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three << two == result
        except ValueError:
            pass

    def test_rshift(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three >> two, 3 >> 2)
        self.assertEqual(3 >> two, 3 >> 2)
        self.assertEqual(three >> 2, 3 >> 2)

    def test_rshift_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three >> two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three >> two == result
        except ValueError:
            pass

    def test_and(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three & two, 3 & 2)
        self.assertEqual(3 & two, 3 & 2)
        self.assertEqual(three & 2, 3 & 2)

    def test_and_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three & two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three & two == result
        except ValueError:
            pass

    def test_xor(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three ^ two, 3 ^ 2)
        self.assertEqual(3 ^ two, 3 ^ 2)
        self.assertEqual(three ^ 2, 3 ^ 2)

    def test_xor_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three ^ two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three ^ two == result
        except ValueError:
            pass

    def test_or(self):
        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy(3)

        self.assertEqual(three | two, 3 | 2)
        self.assertEqual(3 | two, 3 | 2)
        self.assertEqual(three | 2, 3 | 2)

    def test_or_uninitialized_args(self):
        result = object()

        two = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)
        three = wrapt.ObjectProxy(3)

        try:
            assert three | two == result
        except ValueError:
            pass

        two = wrapt.ObjectProxy(2)
        three = wrapt.ObjectProxy.__new__(wrapt.ObjectProxy, None)

        try:
            assert three | two == result
        except ValueError:
            pass

    def test_iadd(self):
        value = wrapt.ObjectProxy(1)
        one = wrapt.ObjectProxy(1)

        value += 1
        self.assertEqual(value, 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value += one
        self.assertEqual(value, 3)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_isub(self):
        value = wrapt.ObjectProxy(1)
        one = wrapt.ObjectProxy(1)

        value -= 1
        self.assertEqual(value, 0)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value -= one
        self.assertEqual(value, -1)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_imul(self):
        value = wrapt.ObjectProxy(2)
        two = wrapt.ObjectProxy(2)

        value *= 2
        self.assertEqual(value, 4)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value *= two
        self.assertEqual(value, 8)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_idiv(self):
        # On Python 2 this will pick up div and on Python
        # 3 it will pick up truediv.

        value = wrapt.ObjectProxy(2)
        two = wrapt.ObjectProxy(2)

        value /= 2
        self.assertEqual(value, 2 / 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value /= two
        self.assertEqual(value, 2 / 2 / 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ifloordiv(self):
        value = wrapt.ObjectProxy(2)
        two = wrapt.ObjectProxy(2)

        value //= 2
        self.assertEqual(value, 2 // 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value //= two
        self.assertEqual(value, 2 // 2 // 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_imod(self):
        value = wrapt.ObjectProxy(10)
        two = wrapt.ObjectProxy(2)

        value %= 2
        self.assertEqual(value, 10 % 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value %= two
        self.assertEqual(value, 10 % 2 % 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ipow(self):
        value = wrapt.ObjectProxy(10)
        two = wrapt.ObjectProxy(2)

        value **= 2
        self.assertEqual(value, 10**2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value **= two
        self.assertEqual(value, 10**2**2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ilshift(self):
        value = wrapt.ObjectProxy(256)
        two = wrapt.ObjectProxy(2)

        value <<= 2
        self.assertEqual(value, 256 << 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value <<= two
        self.assertEqual(value, 256 << 2 << 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_irshift(self):
        value = wrapt.ObjectProxy(2)
        two = wrapt.ObjectProxy(2)

        value >>= 2
        self.assertEqual(value, 2 >> 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value >>= two
        self.assertEqual(value, 2 >> 2 >> 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_iand(self):
        value = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)

        value &= 2
        self.assertEqual(value, 1 & 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value &= two
        self.assertEqual(value, 1 & 2 & 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ixor(self):
        value = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)

        value ^= 2
        self.assertEqual(value, 1 ^ 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value ^= two
        self.assertEqual(value, 1 ^ 2 ^ 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ior(self):
        value = wrapt.ObjectProxy(1)
        two = wrapt.ObjectProxy(2)

        value |= 2
        self.assertEqual(value, 1 | 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

        value |= two
        self.assertEqual(value, 1 | 2 | 2)

        self.assertTrue(isinstance(value, wrapt.ObjectProxy))

    def test_ior_list_self(self):
        value = wrapt.ObjectProxy([])

        try:
            value |= value
        except TypeError:
            pass

    def test_neg(self):
        value = wrapt.ObjectProxy(1)

        self.assertEqual(-value, -1)

    def test_pos(self):
        value = wrapt.ObjectProxy(1)

        self.assertEqual(+value, 1)

    def test_abs(self):
        value = wrapt.ObjectProxy(-1)

        self.assertEqual(abs(value), 1)

    def test_invert(self):
        value = wrapt.ObjectProxy(1)

        self.assertEqual(~value, ~1)

    def test_oct(self):
        value = wrapt.ObjectProxy(20)

        self.assertEqual(oct(value), oct(20))

    def test_hex(self):
        value = wrapt.ObjectProxy(20)

        self.assertEqual(hex(value), hex(20))

    def test_index(self):
        class Class:
            def __index__(self):
                return 1

        value = wrapt.ObjectProxy(Class())
        items = [0, 1, 2]

        self.assertEqual(items[value], items[1])


class TestAsSequenceObjectProxy(unittest.TestCase):

    def test_length(self):
        value = wrapt.ObjectProxy(list(range(3)))

        self.assertEqual(len(value), 3)

    def test_contains(self):
        value = wrapt.ObjectProxy(list(range(3)))

        self.assertTrue(2 in value)
        self.assertFalse(-2 in value)

    def test_getitem(self):
        value = wrapt.ObjectProxy(list(range(3)))

        self.assertEqual(value[1], 1)

    def test_setitem(self):
        value = wrapt.ObjectProxy(list(range(3)))
        value[1] = -1

        self.assertEqual(value[1], -1)

    def test_delitem(self):
        value = wrapt.ObjectProxy(list(range(3)))

        self.assertEqual(len(value), 3)

        del value[1]

        self.assertEqual(len(value), 2)
        self.assertEqual(value[1], 2)

    def test_getslice(self):
        value = wrapt.ObjectProxy(list(range(5)))

        self.assertEqual(value[1:4], [1, 2, 3])

    def test_setslice(self):
        value = wrapt.ObjectProxy(list(range(5)))

        value[1:4] = reversed(value[1:4])

        self.assertEqual(value[1:4], [3, 2, 1])

    def test_delslice(self):
        value = wrapt.ObjectProxy(list(range(5)))

        del value[1:4]

        self.assertEqual(len(value), 2)
        self.assertEqual(value, [0, 4])


class TestAsMappingObjectProxy(unittest.TestCase):

    def test_length(self):
        value = wrapt.ObjectProxy(dict.fromkeys(range(3), False))

        self.assertEqual(len(value), 3)

    def test_contains(self):
        value = wrapt.ObjectProxy(dict.fromkeys(range(3), False))

        self.assertTrue(2 in value)
        self.assertFalse(-2 in value)

    def test_getitem(self):
        value = wrapt.ObjectProxy(dict.fromkeys(range(3), False))

        self.assertEqual(value[1], False)

    def test_setitem(self):
        value = wrapt.ObjectProxy(dict.fromkeys(range(3), False))
        value[1] = True

        self.assertEqual(value[1], True)

    def test_delitem(self):
        value = wrapt.ObjectProxy(dict.fromkeys(range(3), False))

        self.assertEqual(len(value), 3)

        del value[1]

        self.assertEqual(len(value), 2)


class TestObjectRepresentationObjectProxy(unittest.TestCase):

    def test_str(self):
        value = wrapt.ObjectProxy(10)

        self.assertEqual(str(value), str(10))

        value = wrapt.ObjectProxy((10,))

        self.assertEqual(str(value), str((10,)))

        value = wrapt.ObjectProxy([10])

        self.assertEqual(str(value), str([10]))

        value = wrapt.ObjectProxy({10: 10})

        self.assertEqual(str(value), str({10: 10}))

    def test_repr(self):
        number = 10
        value = wrapt.ObjectProxy(number)

        self.assertNotEqual(repr(value).find("ObjectProxy at"), -1)

    def test_format(self):
        value = 1

        proxy = wrapt.ObjectProxy(1)

        self.assertEqual(f"{proxy:0>3}", f"{value:0>3}")


class TestDerivedClassCreation(unittest.TestCase):

    def test_derived_new(self):

        class DerivedObjectProxy(wrapt.ObjectProxy):

            def __new__(cls, wrapped):
                instance = super(DerivedObjectProxy, cls).__new__(cls, wrapped)
                instance.__init__(wrapped)

            def __init__(self, wrapped):
                super(DerivedObjectProxy, self).__init__(wrapped)

        def function():
            pass

        obj = DerivedObjectProxy(function)

    def test_derived_setattr(self):

        class DerivedObjectProxy(wrapt.ObjectProxy):

            def __init__(self, wrapped):
                self._self_attribute = True
                super(DerivedObjectProxy, self).__init__(wrapped)

        def function():
            pass

        obj = DerivedObjectProxy(function)

    def test_derived_missing_init(self):

        class DerivedObjectProxy(wrapt.ObjectProxy):

            def __init__(self, wrapped):
                self.__wrapped__ = wrapped

        def function():
            pass

        obj = DerivedObjectProxy(function)

        self.assertEqual(function, obj)
        self.assertEqual(function, obj.__wrapped__)


class DerivedClassAttributes(unittest.TestCase):

    def test_setup_class_attributes(self):

        def function():
            pass

        class DerivedObjectProxy(wrapt.ObjectProxy):
            pass

        obj = DerivedObjectProxy(function)

        DerivedObjectProxy.ATTRIBUTE = 1

        self.assertEqual(obj.ATTRIBUTE, 1)
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

        del DerivedObjectProxy.ATTRIBUTE

        self.assertFalse(hasattr(DerivedObjectProxy, "ATTRIBUTE"))
        self.assertFalse(hasattr(obj, "ATTRIBUTE"))
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

    def test_override_class_attributes(self):

        def function():
            pass

        class DerivedObjectProxy(wrapt.ObjectProxy):
            ATTRIBUTE = 1

        obj = DerivedObjectProxy(function)

        self.assertEqual(DerivedObjectProxy.ATTRIBUTE, 1)
        self.assertEqual(obj.ATTRIBUTE, 1)

        obj.ATTRIBUTE = 2

        self.assertEqual(DerivedObjectProxy.ATTRIBUTE, 1)

        self.assertEqual(obj.ATTRIBUTE, 2)
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

        del DerivedObjectProxy.ATTRIBUTE

        self.assertFalse(hasattr(DerivedObjectProxy, "ATTRIBUTE"))
        self.assertEqual(obj.ATTRIBUTE, 2)
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

    def test_class_properties(self):

        def function():
            pass

        class DerivedObjectProxy(wrapt.ObjectProxy):
            def __init__(self, wrapped):
                super(DerivedObjectProxy, self).__init__(wrapped)
                self._self_attribute = 1

            @property
            def ATTRIBUTE(self):
                return self._self_attribute

            @ATTRIBUTE.setter
            def ATTRIBUTE(self, value):
                self._self_attribute = value

            @ATTRIBUTE.deleter
            def ATTRIBUTE(self):
                del self._self_attribute

        obj = DerivedObjectProxy(function)

        self.assertEqual(obj.ATTRIBUTE, 1)

        obj.ATTRIBUTE = 2

        self.assertEqual(obj.ATTRIBUTE, 2)
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

        del obj.ATTRIBUTE

        self.assertFalse(hasattr(obj, "ATTRIBUTE"))
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

        obj.ATTRIBUTE = 1

        self.assertEqual(obj.ATTRIBUTE, 1)

        obj.ATTRIBUTE = 2

        self.assertEqual(obj.ATTRIBUTE, 2)
        self.assertFalse(hasattr(function, "ATTRIBUTE"))

        del obj.ATTRIBUTE

        self.assertFalse(hasattr(obj, "ATTRIBUTE"))
        self.assertFalse(hasattr(function, "ATTRIBUTE"))


class OverrideAttributeAccess(unittest.TestCase):

    def test_attr_functions(self):

        def function():
            pass

        proxy = wrapt.ObjectProxy(function)

        self.assertTrue(hasattr(proxy, "__getattr__"))
        self.assertTrue(hasattr(proxy, "__setattr__"))
        self.assertTrue(hasattr(proxy, "__delattr__"))

    def test_override_getattr(self):

        def function():
            pass

        accessed = []

        class DerivedObjectProxy(wrapt.ObjectProxy):
            def __getattr__(self, name):
                accessed.append(name)
                try:
                    __getattr__ = super(DerivedObjectProxy, self).__getattr__
                except AttributeError as e:
                    raise RuntimeError(str(e))
                return __getattr__(name)

        function.attribute = 1

        proxy = DerivedObjectProxy(function)

        self.assertEqual(proxy.attribute, 1)

        self.assertTrue("attribute" in accessed)


class CallableFunction(unittest.TestCase):

    def test_proxy_hasattr_call(self):
        proxy = wrapt.ObjectProxy(None)

        self.assertFalse(hasattr(proxy, "__call__"))

    def test_proxy_getattr_call(self):
        proxy = wrapt.ObjectProxy(None)

        self.assertEqual(getattr(proxy, "__call__", None), None)

    def test_proxy_is_callable(self):
        proxy = wrapt.ObjectProxy(None)

        self.assertFalse(callable(proxy))

    def test_callable_proxy_hasattr_call(self):
        proxy = wrapt.CallableObjectProxy(None)

        self.assertTrue(hasattr(proxy, "__call__"))

    def test_callable_proxy_getattr_call(self):
        proxy = wrapt.CallableObjectProxy(None)

        self.assertTrue(getattr(proxy, "__call__", None), None)

    def test_callable_proxy_is_callable(self):
        proxy = wrapt.CallableObjectProxy(None)

        self.assertTrue(callable(proxy))


class SpecialMethods(unittest.TestCase):

    def test_class_bytes(self):
        class Class:
            def __bytes__(self):
                return b"BYTES"

        instance = Class()

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(bytes(instance), bytes(proxy))

    def test_str_format(self):
        instance = "abcd"

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(format(instance, ""), format(proxy, ""))

    def test_list_reversed(self):
        instance = [1, 2]

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(list(reversed(instance)), list(reversed(proxy)))

    def test_complex(self):
        instance = 1.0 + 2j

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(complex(instance), complex(proxy))

    def test_decimal_complex(self):
        import decimal

        instance = decimal.Decimal(123)

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(complex(instance), complex(proxy))

    def test_fractions_round(self):
        import fractions

        instance = fractions.Fraction("1/2")

        proxy = wrapt.ObjectProxy(instance)

        self.assertEqual(round(instance), round(proxy))
        self.assertEqual(round(instance, 3), round(proxy, 3))
        self.assertEqual(round(instance, ndigits=3), round(proxy, ndigits=3))


class TestArgumentUnpacking(unittest.TestCase):

    def test_self_keyword_argument_on_dict(self):
        # A dict when given self as keyword argument uses it to create item in
        # the dict and no attempt is made to use a positional argument.

        d = wrapt.CallableObjectProxy(dict)(self="self")

        self.assertEqual(d, dict(self="self"))

    def test_self_positional_argument_on_class_init(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        o = Object("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

        o = wrapt.CallableObjectProxy(Object)("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

    def test_self_keyword_argument_on_class_init_1(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.CallableObjectProxy(Object)(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_2(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.CallableObjectProxy(Object)(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_renamed(self):
        class Object:
            def __init__(_self, *args, **kwargs):
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

        o = wrapt.CallableObjectProxy(Object)(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

    def test_self_keyword_argument_on_class_init_overloaded_1(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

        o = wrapt.CallableObjectProxy(Object)(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

    def test_self_keyword_argument_on_class_init_overloaded_2(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.CallableObjectProxy(Object)(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )


class TestArgumentUnpackingPartial(unittest.TestCase):

    def test_self_keyword_argument_on_dict_1(self):
        # A dict when given self as keyword argument uses it to create item in
        # the dict and no attempt is made to use a positional argument.

        wrapper = wrapt.partial(dict, arg1="arg1")

        d = wrapper(self="self")

        self.assertEqual(d, dict(self="self", arg1="arg1"))

    def test_self_keyword_argument_on_dict_2(self):
        # A dict when given self as keyword argument uses it to create item in
        # the dict and no attempt is made to use a positional argument.

        wrapper = wrapt.partial(dict, self="self")

        d = wrapper(arg1="arg1")

        self.assertEqual(d, dict(self="self", arg1="arg1"))

    def test_self_positional_argument_on_class_init_1(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        o = Object("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

        wrapper = wrapt.partial(Object, "arg1")

        o = wrapper()

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

    def test_self_positional_argument_on_class_init_2(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        o = Object("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

        wrapper = wrapt.partial(Object)

        o = wrapper("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

    def test_self_keyword_argument_on_class_init_1a(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object, self="self")

        with self.assertRaises(TypeError) as e:
            o = wrapper()

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_1b(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object)

        with self.assertRaises(TypeError) as e:
            o = wrapper(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_2a(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object, arg1="arg1", self="self")

        with self.assertRaises(TypeError) as e:
            o = wrapper()

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_2b(self):
        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object)

        with self.assertRaises(TypeError) as e:
            o = wrapper(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_renamed_1(self):
        class Object:
            def __init__(_self, *args, **kwargs):
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

        wrapper = wrapt.partial(Object, self="self")

        o = wrapper()

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

    def test_self_keyword_argument_on_class_init_renamed_2(self):
        class Object:
            def __init__(_self, *args, **kwargs):
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

        wrapper = wrapt.partial(Object)

        o = wrapper(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

    def test_self_keyword_argument_on_class_init_overloaded_1a(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

        wrapper = wrapt.partial(Object, self="self")

        o = wrapper()

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

    def test_self_keyword_argument_on_class_init_overloaded_1b(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

        wrapper = wrapt.partial(Object)

        o = wrapper(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

    def test_self_keyword_argument_on_class_init_overloaded_2a(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object, _self="self")

        with self.assertRaises(TypeError) as e:
            o = wrapper()

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_overloaded_2b(self):
        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )

        wrapper = wrapt.partial(Object)

        with self.assertRaises(TypeError) as e:
            o = wrapper(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )


class TestArgumentUnpackingWrapperBase(unittest.TestCase):

    def test_self_keyword_argument_on_dict(self):
        # A dict when given self as keyword argument uses it to create item in
        # the dict and no attempt is made to use a positional argument.

        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        d = wrapt.FunctionWrapper(dict, wrapper)(self="self")

        self.assertEqual(d, dict(self="self"))

    def test_self_positional_argument_on_class_init(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        o = Object("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

        o = wrapt.FunctionWrapper(Object, wrapper)("arg1")

        self.assertEqual(o._args, ("arg1",))
        self.assertEqual(o._kwargs, {})

    def test_self_keyword_argument_on_class_init_1(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.FunctionWrapper(Object, wrapper)(self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_2(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(self, *args, **kwargs):
                self._args = args
                self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.FunctionWrapper(Object, wrapper)(arg1="arg1", self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument 'self'.*",
                str(e.exception),
            ),
            None,
        )

    def test_self_keyword_argument_on_class_init_renamed(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(_self, *args, **kwargs):
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

        o = wrapt.FunctionWrapper(Object, wrapper)(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, dict(self="self"))

    def test_self_keyword_argument_on_class_init_overloaded_1(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        o = Object(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

        o = wrapt.FunctionWrapper(Object, wrapper)(self="self")

        self.assertEqual(o._args, ())
        self.assertEqual(o._kwargs, {})
        self.assertEqual(o._self, "self")

    def test_self_keyword_argument_on_class_init_overloaded_2(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def __init__(_self, self, *args, **kwargs):
                _self._self = self
                _self._args = args
                _self._kwargs = kwargs

        with self.assertRaises(TypeError) as e:
            Object(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )

        with self.assertRaises(TypeError) as e:
            wrapt.FunctionWrapper(Object, wrapper)(_self="self")

        self.assertNotEqual(
            re.match(
                ".*got multiple values for (keyword )?argument '_self'.*",
                str(e.exception),
            ),
            None,
        )


class TestArgumentUnpackingBoundFunctionWrapper(unittest.TestCase):

    def test_self_keyword_argument_on_classmethod(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            @classmethod
            def function(cls, self, *args, **kwargs):
                return self, args, kwargs

            function = wrapt.FunctionWrapper(function, wrapper)

        result = Object().function(self="self")

        self.assertEqual(result, ("self", (), {}))

    def test_self_keyword_argument_on_instancemethod(self):
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            def function(_self, self, *args, **kwargs):
                return self, args, kwargs

            function = wrapt.FunctionWrapper(function, wrapper)

        result = Object().function(self="self")

        self.assertEqual(result, ("self", (), {}))


class TestArgumentUnpackingDecorator(unittest.TestCase):

    def test_self_keyword_argument_on_function(self):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        @wrapper
        def function(self, *args, **kwargs):
            return self, args, kwargs

        result = function(self="self")

        self.assertEqual(result, ("self", (), {}))

        result = function("self")

        self.assertEqual(result, ("self", (), {}))

    def test_self_keyword_argument_on_classmethod(self):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            @wrapper
            @classmethod
            def function(cls, self, *args, **kwargs):
                return self, args, kwargs

        result = Object().function(self="self")

        self.assertEqual(result, ("self", (), {}))

        result = Object().function("self", arg1="arg1")

        self.assertEqual(result, ("self", (), dict(arg1="arg1")))

    def test_self_keyword_argument_on_instancemethod(self):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            return wrapped(*args, **kwargs)

        class Object:
            @wrapper
            def function(_self, self, *args, **kwargs):
                return self, args, kwargs

        result = Object().function(self="self", arg1="arg1")

        self.assertEqual(result, ("self", (), dict(arg1="arg1")))

        result = Object().function("self", arg1="arg1")

        self.assertEqual(result, ("self", (), dict(arg1="arg1")))


class TestOverridingSpecialAttributes(unittest.TestCase):

    def test_overriding_class_attribute(self):
        class Object1:
            pass

        class Object2(Object1):
            pass

        o1 = Object1()

        self.assertEqual(o1.__class__, type(o1))

        o2 = Object2()

        self.assertEqual(o2.__class__, type(o2))

        o2.__class__ = type(o1)

        self.assertEqual(o2.__class__, type(o1))


class TestClassGetItem(unittest.TestCase):

    def test_class_getitem(self):
        class Meta(type):
            def __getitem__(cls, key):
                return key

        class Object(metaclass=Meta):
            pass

        self.assertEqual(Object["key"], "key")

        proxy = wrapt.ObjectProxy(Object)

        self.assertEqual(proxy["key"], "key")


if __name__ == "__main__":
    unittest.main()



"""
This example demonstrates the correct usage of the FunctionWrapper type.

It covers the following cases:
- Wrapping a function
- Wrapping a method of a class
- Wrapping a class method
- Wrapping a static method
- Wrapping an instance method of a class
- Wrapping a class method of an instance
- Wrapping a static method of an instance
- Wrapping a callable class
- Wrapping a callable class instance

These should all pass mypy type checking.
"""

from typing import Any, Callable, ParamSpec, TypeVar

from wrapt import FunctionWrapper

P = ParamSpec("P")
R = TypeVar("R", covariant=True)


def function(x: int, y: int = 0) -> int:
    """A simple function to be wrapped."""
    return x + y


class ExampleClass:
    """A class with methods to be wrapped."""

    def __init__(self, value: int) -> None:
        self.value = value

    def __call__(self, x: int, y: int = 0) -> int:
        return x + y

    def instance_method(self, x: int, y: int = 0) -> int:
        return x + y

    @classmethod
    def class_method(cls, x: int, y: int = 0) -> int:
        return x + y

    @staticmethod
    def static_method(x: int, y: int = 0) -> int:
        return x + y


def standard_wrapper(
    wrapped: Callable[P, R],
    instance: Any,
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> R:
    try:
        print(f"Before calling {wrapped.__name__}")
        return wrapped(*args, **kwargs)
    finally:
        print(f"After calling {wrapped.__name__}")


wrapped_function = FunctionWrapper(function, standard_wrapper)

wrapped_method = FunctionWrapper(ExampleClass.instance_method, standard_wrapper)
wrapped_classmethod = FunctionWrapper(ExampleClass.class_method, standard_wrapper)
wrapped_staticmethod = FunctionWrapper(ExampleClass.static_method, standard_wrapper)


wrapped_method_instance = FunctionWrapper(
    ExampleClass(0).instance_method, standard_wrapper
)
wrapped_classmethod_instance = FunctionWrapper(
    ExampleClass(0).class_method, standard_wrapper
)
wrapped_staticmethod_instance = FunctionWrapper(
    ExampleClass(0).static_method, standard_wrapper
)

wrapped_callable_class = FunctionWrapper(ExampleClass, standard_wrapper)
wrapped_callable_object = FunctionWrapper(ExampleClass(0), standard_wrapper)

wrapped_function(1)
wrapped_method(ExampleClass(0), 3)
wrapped_classmethod(4)
wrapped_staticmethod(5)
wrapped_method_instance(6)
wrapped_classmethod_instance(7)
wrapped_staticmethod_instance(8)

wrapped_callable_class(9)
wrapped_callable_object(10)



import sys
import unittest
from inspect import getfullargspec

from wrapt.arguments import formatargspec


class TestFormatargspec38(unittest.TestCase):

    def assertFormatEqual(self, func, ref):
        formatted = formatargspec(*getfullargspec(func))
        self.assertEqual(formatted, ref)

    def test_formatargspec(self):
        # examples from https://www.python.org/dev/peps/pep-0570/
        def name1(p1, p2, /, p_or_kw, *, kw):
            pass

        self.assertFormatEqual(name1, "(p1, p2, p_or_kw, *, kw)")

        def name2(p1, p2=None, /, p_or_kw=None, *, kw):
            pass

        self.assertFormatEqual(name2, "(p1, p2=None, p_or_kw=None, *, kw)")

        def name3(p1, p2=None, /, *, kw):
            pass

        self.assertFormatEqual(name3, "(p1, p2=None, *, kw)")

        def name4(p1, p2=None, /):
            pass

        self.assertFormatEqual(name4, "(p1, p2=None)")

        def name5(p1, p2, /, p_or_kw):
            pass

        self.assertFormatEqual(name5, "(p1, p2, p_or_kw)")

        def name6(p1, p2, /):
            pass

        self.assertFormatEqual(name6, "(p1, p2)")

        def name7(p_or_kw, *, kw):
            pass

        self.assertFormatEqual(name7, "(p_or_kw, *, kw)")

        def name8(*, kw):
            pass

        self.assertFormatEqual(name8, "(*, kw)")
